@@ -18,13 +18,12 @@ execution to date.
 import collections
 import dataclasses
 import inspect
-from collections.abc import Iterable, Sequence
-from copy import copy
+from collections.abc import Callable, Iterable, Sequence
 from dataclasses import dataclass, field
 from functools import lru_cache
 from io import StringIO
 from time import perf_counter
-from typing import Any, Callable, ClassVar, Optional, TypeVar, Union, overload
+from typing import Any, ClassVar, TypeVar, overload
 from unittest import TestCase
 
 from hypothesis import strategies as st
@@ -42,7 +41,7 @@ from hypothesis.internal.conjecture import utils as cu
 from hypothesis.internal.conjecture.engine import BUFFER_SIZE
 from hypothesis.internal.conjecture.junkdrawer import gc_cumulative_time
 from hypothesis.internal.healthcheck import fail_health_check
-from hypothesis.internal.observability import TESTCASE_CALLBACKS
+from hypothesis.internal.observability import observability_enabled
 from hypothesis.internal.reflection import (
     function_digest,
     get_pretty_function_description,
@@ -113,9 +112,9 @@ def get_state_machine_test(state_machine_factory, *, settings=None, _min_steps=0
 
     @settings
     @given(st.data())
-    def run_state_machine(factory, data):
+    def run_state_machine(data):
         cd = data.conjecture_data
-        machine = factory()
+        machine: RuleBasedStateMachine = state_machine_factory()
         check_type(RuleBasedStateMachine, machine, "state_machine_factory()")
         cd.hypothesis_runner = machine
         machine._observability_predicates = cd._observability_predicates  # alias
@@ -128,7 +127,7 @@ def get_state_machine_test(state_machine_factory, *, settings=None, _min_steps=0
         def output(s):
             if print_steps:
                 report(s)
-            if TESTCASE_CALLBACKS:
+            if observability_enabled():
                 cd._stateful_repr_parts.append(s)
 
         try:
@@ -181,7 +180,7 @@ def get_state_machine_test(state_machine_factory, *, settings=None, _min_steps=0
                 # _add_results_to_targets, to avoid printing arguments which are also
                 # a return value using the variable name they are assigned to.
                 # See https://github.com/HypothesisWorks/hypothesis/issues/2341
-                if print_steps or TESTCASE_CALLBACKS:
+                if print_steps or observability_enabled():
                     data_to_print = {
                         k: machine._pretty_print(v) for k, v in data.items()
                     }
@@ -218,7 +217,7 @@ def get_state_machine_test(state_machine_factory, *, settings=None, _min_steps=0
                             HealthCheck.return_value,
                         )
                 finally:
-                    if print_steps or TESTCASE_CALLBACKS:
+                    if print_steps or observability_enabled():
                         # 'result' is only used if the step has target bundles.
                         # If it does, and the result is a 'MultipleResult',
                         # then 'print_step' prints a multi-variable assignment.
@@ -255,7 +254,7 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None, _min_step
     state_machine_test = get_state_machine_test(
         state_machine_factory, settings=settings, _min_steps=_min_steps
     )
-    state_machine_test(state_machine_factory)
+    state_machine_test()
 
 
 class StateMachineMeta(type):
@@ -270,6 +269,13 @@ class StateMachineMeta(type):
         return super().__setattr__(name, value)
 
 
+@dataclass
+class _SetupState:
+    rules: list["Rule"]
+    invariants: list["Invariant"]
+    initializers: list["Rule"]
+
+
 class RuleBasedStateMachine(metaclass=StateMachineMeta):
     """A RuleBasedStateMachine gives you a structured way to define state machines.
 
@@ -281,23 +287,14 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
     At any given point a random applicable rule will be executed.
     """
 
-    _rules_per_class: ClassVar[dict[type, list["Rule"]]] = {}
-    _invariants_per_class: ClassVar[dict[type, list["Invariant"]]] = {}
-    _initializers_per_class: ClassVar[dict[type, list["Rule"]]] = {}
+    _setup_state_per_class: ClassVar[dict[type, _SetupState]] = {}
 
     def __init__(self) -> None:
-        if not self.rules():
-            raise InvalidDefinition(f"Type {type(self).__name__} defines no rules")
-        self.bundles: dict[str, list] = {}
-        self.names_counters: collections.Counter = collections.Counter()
-        self.names_list: list[str] = []
-        self.names_to_values: dict[str, Any] = {}
-        self.__stream = StringIO()
-        self.__printer = RepresentationPrinter(
-            self.__stream, context=_current_build_context.value
-        )
-        self._initialize_rules_to_run = copy(self.initialize_rules())
-        self._rules_strategy = RuleStrategy(self)
+        setup_state = self.setup_state()
+        if not setup_state.rules:
+            raise InvalidDefinition(
+                f"State machine {type(self).__name__} defines no rules"
+            )
 
         if isinstance(s := vars(type(self)).get("settings"), Settings):
             tname = type(self).__name__
@@ -308,6 +305,21 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
                 f"on the {tname} class."
             )
 
+        self.rules = setup_state.rules
+        self.invariants = setup_state.invariants
+        # copy since we pop from this as we run initialize rules.
+        self._initialize_rules_to_run = setup_state.initializers.copy()
+
+        self.bundles: dict[str, list] = {}
+        self.names_counters: collections.Counter = collections.Counter()
+        self.names_list: list[str] = []
+        self.names_to_values: dict[str, Any] = {}
+        self.__stream = StringIO()
+        self.__printer = RepresentationPrinter(
+            self.__stream, context=_current_build_context.value
+        )
+        self._rules_strategy = RuleStrategy(self)
+
     def _pretty_print(self, value):
         if isinstance(value, VarReference):
             return value.name
@@ -342,49 +354,44 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
         return self.bundles.setdefault(name, [])
 
     @classmethod
-    def initialize_rules(cls) -> list["Rule"]:
+    def setup_state(cls):
         try:
-            return cls._initializers_per_class[cls]
+            return cls._setup_state_per_class[cls]
         except KeyError:
             pass
 
-        initializers = []
-        for _, v in inspect.getmembers(cls):
-            r = getattr(v, INITIALIZE_RULE_MARKER, None)
-            if r is not None:
-                initializers.append(r)
-        cls._initializers_per_class[cls] = initializers
-        return initializers
+        rules: list[Rule] = []
+        initializers: list[Rule] = []
+        invariants: list[Invariant] = []
 
-    @classmethod
-    def rules(cls) -> list["Rule"]:
-        try:
-            return cls._rules_per_class[cls]
-        except KeyError:
-            pass
-
-        rules = []
-        for _, v in inspect.getmembers(cls):
-            r = getattr(v, RULE_MARKER, None)
-            if r is not None:
-                rules.append(r)
-        cls._rules_per_class[cls] = rules
-        return rules
-
-    @classmethod
-    def invariants(cls) -> list["Invariant"]:
-        try:
-            return cls._invariants_per_class[cls]
-        except KeyError:
-            pass
-
-        invariants = []
-        for _, v in inspect.getmembers(cls):
-            invariant = getattr(v, INVARIANT_MARKER, None)
+        for _name, f in inspect.getmembers(cls):
+            rule = getattr(f, RULE_MARKER, None)
+            initializer = getattr(f, INITIALIZE_RULE_MARKER, None)
+            invariant = getattr(f, INVARIANT_MARKER, None)
+            if rule is not None:
+                rules.append(rule)
+            if initializer is not None:
+                initializers.append(initializer)
             if invariant is not None:
                 invariants.append(invariant)
-        cls._invariants_per_class[cls] = invariants
-        return invariants
+
+            if (
+                getattr(f, PRECONDITIONS_MARKER, None) is not None
+                and rule is None
+                and invariant is None
+            ):
+                raise InvalidDefinition(
+                    f"{_rule_qualname(f)} has been decorated with @precondition, "
+                    "but not @rule (or @invariant), which is not allowed. A "
+                    "precondition must be combined with a rule or an invariant, "
+                    "since it has no effect alone."
+                )
+
+        state = _SetupState(
+            rules=rules, initializers=initializers, invariants=invariants
+        )
+        cls._setup_state_per_class[cls] = state
+        return state
 
     def _repr_step(self, rule: "Rule", data: Any, result: Any) -> str:
         output_assignment = ""
@@ -435,7 +442,7 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
                 self.bundles.setdefault(target, []).append(VarReference(name))
 
     def check_invariants(self, settings, output, runtimes):
-        for invar in self.invariants():
+        for invar in self.invariants:
             if self._initialize_rules_to_run and not invar.check_during_init:
                 continue
             if not all(precond(self) for precond in invar.preconditions):
@@ -444,7 +451,7 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
             if (
                 current_build_context().is_final
                 or settings.verbosity >= Verbosity.debug
-                or TESTCASE_CALLBACKS
+                or observability_enabled()
             ):
                 output(f"state.{name}()")
             start = perf_counter()
@@ -492,6 +499,8 @@ class Rule:
     arguments: Any
     preconditions: Any
     bundles: tuple["Bundle", ...] = field(init=False)
+    _cached_hash: int | None = field(init=False, default=None)
+    _cached_repr: str | None = field(init=False, default=None)
 
     def __post_init__(self):
         self.arguments_strategies = {}
@@ -506,13 +515,30 @@ class Rule:
         self.bundles = tuple(bundles)
 
     def __repr__(self) -> str:
-        bits = [
-            f"{field.name}="
-            f"{get_pretty_function_description(getattr(self, field.name))}"
-            for field in dataclasses.fields(self)
-            if getattr(self, field.name)
-        ]
-        return f"{self.__class__.__name__}({', '.join(bits)})"
+        if self._cached_repr is None:
+            bits = [
+                f"{field.name}="
+                f"{get_pretty_function_description(getattr(self, field.name))}"
+                for field in dataclasses.fields(self)
+                if getattr(self, field.name)
+            ]
+            self._cached_repr = f"{self.__class__.__name__}({', '.join(bits)})"
+        return self._cached_repr
+
+    def __hash__(self):
+        # sampled_from uses hash in calc_label, and we want this to be fast when
+        # sampling stateful rules, so we cache here.
+        if self._cached_hash is None:
+            self._cached_hash = hash(
+                (
+                    self.targets,
+                    self.function,
+                    tuple(self.arguments.items()),
+                    self.preconditions,
+                    self.bundles,
+                )
+            )
+        return self._cached_hash
 
 
 self_strategy = st.runner()
@@ -520,6 +546,7 @@ self_strategy = st.runner()
 
 class BundleReferenceStrategy(SearchStrategy):
     def __init__(self, name: str, *, consume: bool = False):
+        super().__init__()
         self.name = name
         self.consume = consume
 
@@ -563,6 +590,7 @@ class Bundle(SearchStrategy[Ex]):
     def __init__(
         self, name: str, *, consume: bool = False, draw_references: bool = True
     ) -> None:
+        super().__init__()
         self.name = name
         self.__reference_strategy = BundleReferenceStrategy(name, consume=consume)
         self.draw_references = draw_references
@@ -582,12 +610,12 @@ class Bundle(SearchStrategy[Ex]):
         # We assume that a bundle will grow over time
         return False
 
-    def _available(self, data):
+    def is_currently_empty(self, data):
         # ``self_strategy`` is an instance of the ``st.runner()`` strategy.
         # Hence drawing from it only returns the current state machine without
-        # modifying the underlying buffer.
+        # modifying the underlying choice sequence.
         machine = data.draw(self_strategy)
-        return bool(machine.bundle(self.name))
+        return not bool(machine.bundle(self.name))
 
     def flatmap(self, expand):
         if self.draw_references:
@@ -598,6 +626,14 @@ class Bundle(SearchStrategy[Ex]):
             ).flatmap(expand)
         return super().flatmap(expand)
 
+    def __hash__(self):
+        # Making this hashable means we hit the fast path of "everything is
+        # hashable" in st.sampled_from label calculation when sampling which rule
+        # to invoke next.
+
+        # Mix in "Bundle" for collision resistance
+        return hash(("Bundle", self.name))
+
 
 class BundleConsumer(Bundle[Ex]):
     def __init__(self, bundle: Bundle[Ex]) -> None:
@@ -683,10 +719,20 @@ PRECONDITIONS_MARKER = "hypothesis_stateful_preconditions"
 INVARIANT_MARKER = "hypothesis_stateful_invariant"
 
 
-_RuleType = Callable[..., Union[MultipleResults[Ex], Ex]]
+_RuleType = Callable[..., MultipleResults[Ex] | Ex]
 _RuleWrapper = Callable[[_RuleType[Ex]], _RuleType[Ex]]
 
 
+def _rule_qualname(f: Any) -> str:
+    # we define rules / invariants / initializes inside of wrapper functions, which
+    # makes f.__qualname__ look like:
+    #   test_precondition.<locals>.BadStateMachine.has_precondition_but_no_rule
+    # which is not ideal. This function returns just
+    #   BadStateMachine.has_precondition_but_no_rule
+    # instead.
+    return f.__qualname__.rsplit("<locals>.")[-1]
+
+
 # We cannot exclude `target` or `targets` from any of these signatures because
 # otherwise they would be matched against the `kwargs`, either leading to
 # overlapping overloads of incompatible return types, or a concrete
@@ -741,10 +787,10 @@ def rule(
 
 def rule(
     *,
-    targets: Union[Sequence[Bundle[Ex]], _OmittedArgument] = (),
-    target: Optional[Bundle[Ex]] = None,
+    targets: Sequence[Bundle[Ex]] | _OmittedArgument = (),
+    target: Bundle[Ex] | None = None,
     **kwargs: SearchStrategy,
-) -> Union[_RuleWrapper[Ex], Callable[[Callable[..., None]], Callable[..., None]]]:
+) -> _RuleWrapper[Ex] | Callable[[Callable[..., None]], Callable[..., None]]:
     """Decorator for RuleBasedStateMachine. Any Bundle present in ``target`` or
     ``targets`` will define where the end result of this function should go. If
     both are empty then the end result will be discarded.
@@ -771,15 +817,23 @@ def rule(
     def accept(f):
         if getattr(f, INVARIANT_MARKER, None):
             raise InvalidDefinition(
-                "A function cannot be used for both a rule and an invariant.",
-                Settings.default,
+                f"{_rule_qualname(f)} is used with both @rule and @invariant, "
+                "which is not allowed. A function may be either a rule or an "
+                "invariant, but not both."
             )
         existing_rule = getattr(f, RULE_MARKER, None)
         existing_initialize_rule = getattr(f, INITIALIZE_RULE_MARKER, None)
-        if existing_rule is not None or existing_initialize_rule is not None:
+        if existing_rule is not None:
             raise InvalidDefinition(
-                "A function cannot be used for two distinct rules. ", Settings.default
+                f"{_rule_qualname(f)} has been decorated with @rule twice, which is "
+                "not allowed."
             )
+        if existing_initialize_rule is not None:
+            raise InvalidDefinition(
+                f"{_rule_qualname(f)} has been decorated with both @rule and "
+                "@initialize, which is not allowed."
+            )
+
         preconditions = getattr(f, PRECONDITIONS_MARKER, ())
         rule = Rule(
             targets=converted_targets,
@@ -828,10 +882,10 @@ def initialize(
 
 def initialize(
     *,
-    targets: Union[Sequence[Bundle[Ex]], _OmittedArgument] = (),
-    target: Optional[Bundle[Ex]] = None,
+    targets: Sequence[Bundle[Ex]] | _OmittedArgument = (),
+    target: Bundle[Ex] | None = None,
     **kwargs: SearchStrategy,
-) -> Union[_RuleWrapper[Ex], Callable[[Callable[..., None]], Callable[..., None]]]:
+) -> _RuleWrapper[Ex] | Callable[[Callable[..., None]], Callable[..., None]]:
     """Decorator for RuleBasedStateMachine.
 
     An initialize decorator behaves like a rule, but all ``@initialize()`` decorated
@@ -847,19 +901,28 @@ def initialize(
     def accept(f):
         if getattr(f, INVARIANT_MARKER, None):
             raise InvalidDefinition(
-                "A function cannot be used for both a rule and an invariant.",
-                Settings.default,
+                f"{_rule_qualname(f)} is used with both @initialize and @invariant, "
+                "which is not allowed. A function may be either an initialization "
+                "rule or an invariant, but not both."
             )
         existing_rule = getattr(f, RULE_MARKER, None)
         existing_initialize_rule = getattr(f, INITIALIZE_RULE_MARKER, None)
-        if existing_rule is not None or existing_initialize_rule is not None:
+        if existing_rule is not None:
             raise InvalidDefinition(
-                "A function cannot be used for two distinct rules. ", Settings.default
+                f"{_rule_qualname(f)} has been decorated with both @rule and "
+                "@initialize, which is not allowed."
+            )
+        if existing_initialize_rule is not None:
+            raise InvalidDefinition(
+                f"{_rule_qualname(f)} has been decorated with @initialize twice, "
+                "which is not allowed."
             )
         preconditions = getattr(f, PRECONDITIONS_MARKER, ())
         if preconditions:
             raise InvalidDefinition(
-                "An initialization rule cannot have a precondition. ", Settings.default
+                f"{_rule_qualname(f)} has been decorated with both @initialize and "
+                "@precondition, which is not allowed. An initialization rule "
+                "runs unconditionally and may not have a precondition."
             )
         rule = Rule(
             targets=converted_targets,
@@ -916,7 +979,9 @@ def precondition(precond: Callable[[Any], bool]) -> Callable[[TestFunc], TestFun
         existing_initialize_rule = getattr(f, INITIALIZE_RULE_MARKER, None)
         if existing_initialize_rule is not None:
             raise InvalidDefinition(
-                "An initialization rule cannot have a precondition. ", Settings.default
+                f"{_rule_qualname(f)} has been decorated with both @initialize and "
+                "@precondition, which is not allowed. An initialization rule "
+                "runs unconditionally and may not have a precondition."
             )
 
         rule = getattr(f, RULE_MARKER, None)
@@ -984,14 +1049,14 @@ def invariant(*, check_during_init: bool = False) -> Callable[[TestFunc], TestFu
     def accept(f):
         if getattr(f, RULE_MARKER, None) or getattr(f, INITIALIZE_RULE_MARKER, None):
             raise InvalidDefinition(
-                "A function cannot be used for both a rule and an invariant.",
-                Settings.default,
+                f"{_rule_qualname(f)} has been decorated with both @invariant and "
+                "@rule, which is not allowed."
             )
         existing_invariant = getattr(f, INVARIANT_MARKER, None)
         if existing_invariant is not None:
             raise InvalidDefinition(
-                "A function cannot be used for two distinct invariants.",
-                Settings.default,
+                f"{_rule_qualname(f)} has been decorated with @invariant twice, "
+                "which is not allowed."
             )
         preconditions = getattr(f, PRECONDITIONS_MARKER, ())
         invar = Invariant(
@@ -1014,10 +1079,10 @@ LOOP_LABEL = cu.calc_label_from_name("RuleStrategy loop iteration")
 
 
 class RuleStrategy(SearchStrategy):
-    def __init__(self, machine):
+    def __init__(self, machine: RuleBasedStateMachine) -> None:
         super().__init__()
         self.machine = machine
-        self.rules = list(machine.rules())
+        self.rules = machine.rules.copy()
 
         self.enabled_rules_strategy = st.shared(
             FeatureStrategy(at_least_one_of={r.function.__name__ for r in self.rules}),
@@ -1038,6 +1103,7 @@ class RuleStrategy(SearchStrategy):
                 rule.function.__name__,
             )
         )
+        self.rules_strategy = st.sampled_from(self.rules)
 
     def __repr__(self):
         return f"{self.__class__.__name__}(machine={self.machine.__class__.__name__}({{...}}))"
@@ -1062,7 +1128,7 @@ class RuleStrategy(SearchStrategy):
             # be artificially large.
             return self.is_valid(r) and feature_flags.is_enabled(r.function.__name__)
 
-        rule = data.draw(st.sampled_from(self.rules).filter(rule_is_enabled))
+        rule = data.draw(self.rules_strategy.filter(rule_is_enabled))
 
         arguments = {}
         for k, strat in rule.arguments_strategies.items():
